@@ -30,10 +30,11 @@
 #include "mdadm.h"
 #include	"md_u.h"
 #include	"md_p.h"
+#include	<ctype.h>
 
 int Create(struct supertype *st, char *mddev, int mdfd,
 	   int chunk, int level, int layout, unsigned long long size, int raiddisks, int sparedisks,
-	   char *name, char *homehost,
+	   char *name, char *homehost, int *uuid,
 	   int subdevs, mddev_dev_t devlist,
 	   int runstop, int verbose, int force, int assume_clean,
 	   char *bitmap_file, int bitmap_chunk, int write_behind, int delay)
@@ -72,7 +73,7 @@ int Create(struct supertype *st, char *mddev, int mdfd,
 	unsigned long long bitmapsize;
 
 	mdu_array_info_t array;
-	int major = BITMAP_MAJOR_HI;
+	int major_num = BITMAP_MAJOR_HI;
 
 	memset(&array, 0, sizeof(array));
 
@@ -80,17 +81,21 @@ int Create(struct supertype *st, char *mddev, int mdfd,
 	if (vers < 9000) {
 		fprintf(stderr, Name ": Create requires md driver version 0.90.0 or later\n");
 		return 1;
+	} else {
+		mdu_array_info_t inf;
+		memset(&inf, 0, sizeof(inf));
+		ioctl(mdfd, GET_ARRAY_INFO, &inf);
+		if (inf.working_disks != 0) {
+			fprintf(stderr, Name ": another array by this name"
+				" is already running.\n");
+			return 1;
+		}
 	}
 	if (level == UnSet) {
 		fprintf(stderr,
 			Name ": a RAID level is needed to create an array.\n");
 		return 1;
 	}
-	if (raiddisks < 1) {
-		fprintf(stderr,
-			Name ": a number of --raid-devices must be given to create an array\n");
-		return 1;
-	}
 	if (raiddisks < 4 && level == 6) {
 		fprintf(stderr,
 			Name ": at least 4 raid-devices needed for level 6\n");
@@ -106,6 +111,11 @@ int Create(struct supertype *st, char *mddev, int mdfd,
 			Name ": at least 2 raid-devices needed for level 4 or 5\n");
 		return 1;
 	}
+	if (level <= 0 && sparedisks) {
+		fprintf(stderr,
+			Name ": This level does not support spare devices\n");
+		return 1;
+	}
 	if (subdevs > raiddisks+sparedisks) {
 		fprintf(stderr, Name ": You have listed more devices (%d) than are in the array(%d)!\n", subdevs, raiddisks+sparedisks);
 		return 1;
@@ -114,6 +124,11 @@ int Create(struct supertype *st, char *mddev, int mdfd,
 		fprintf(stderr, Name ": You haven't given enough devices (real or missing) to create this array\n");
 		return 1;
 	}
+	if (bitmap_file && level <= 0) {
+		fprintf(stderr, Name ": bitmaps not meaningful with level %s\n",
+			map_num(pers, level)?:"given");
+		return 1;
+	}
 
 	/* now set some defaults */
 	if (layout == UnSet)
@@ -157,20 +172,24 @@ int Create(struct supertype *st, char *mddev, int mdfd,
 	case 10:
 	case 6:
 	case 0:
-	case -1: /* linear */
+	case LEVEL_LINEAR: /* linear */
 		if (chunk == 0) {
 			chunk = 64;
 			if (verbose > 0)
 				fprintf(stderr, Name ": chunk size defaults to 64K\n");
 		}
 		break;
-	default: /* raid1, multipath */
+	case 1:
+	case LEVEL_MULTIPATH:
 		if (chunk) {
 			chunk = 0;
 			if (verbose > 0)
 				fprintf(stderr, Name ": chunk size ignored for this level\n");
 		}
 		break;
+	default:
+		fprintf(stderr, Name ": unknown level %d\n", level);
+		return 1;
 	}
 
 	/* now look at the subdevs */
@@ -179,7 +198,6 @@ int Create(struct supertype *st, char *mddev, int mdfd,
 	dnum = 0;
 	for (dv=devlist; dv; dv=dv->next, dnum++) {
 		char *dname = dv->devname;
-		unsigned long dsize;
 		unsigned long long ldsize, freesize;
 		int fd;
 		if (strcasecmp(dname, "missing")==0) {
@@ -198,21 +216,15 @@ int Create(struct supertype *st, char *mddev, int mdfd,
 			fail=1;
 			continue;
 		}
-#ifdef BLKGETSIZE64
-		if (ioctl(fd, BLKGETSIZE64, &ldsize)==0)
-			;
-		else
-#endif
-		if (ioctl(fd, BLKGETSIZE, &dsize)) {
-			fprintf(stderr, Name ": Cannot get size of %s: %s\n",
-				dname, strerror(errno));
+		if (!get_dev_size(fd, dname, &ldsize)) {
 			fail = 1;
 			close(fd);
 			continue;
 		}
-		else {
-			ldsize = dsize;
-			ldsize <<= 9;
+		if (st == NULL) {
+			struct createinfo *ci = conf_get_create_info();
+			if (ci)
+				st = ci->supertype;
 		}
 		if (st == NULL) {
 			/* Need to choose a default metadata, which is different
@@ -231,7 +243,8 @@ int Create(struct supertype *st, char *mddev, int mdfd,
 			}
 			if (st->ss->major != 0 ||
 			    st->minor_version != 90)
-				fprintf(stderr, Name ": Defaulting to verion %d.%d metadata\n",
+				fprintf(stderr, Name ": Defaulting to version"
+					" %d.%d metadata\n",
 					st->ss->major,
 					st->minor_version);
 		}
@@ -311,12 +324,13 @@ int Create(struct supertype *st, char *mddev, int mdfd,
 		}
 	}
 
-	/* If this is  raid5, we want to configure the last active slot
+	/* If this is raid4/5, we want to configure the last active slot
 	 * as missing, so that a reconstruct happens (faster than re-parity)
 	 * FIX: Can we do this for raid6 as well?
 	 */
 	if (assume_clean==0 && force == 0 && first_missing >= raiddisks) {
 		switch ( level ) {
+		case 4:
 		case 5:
 			insert_point = raiddisks-1;
 			sparedisks++;
@@ -327,6 +341,12 @@ int Create(struct supertype *st, char *mddev, int mdfd,
 			break;
 		}
 	}
+
+	if (level <= 0 && first_missing != subdevs * 2) {
+		fprintf(stderr,
+			Name ": This level does not support missing devices\n");
+		return 1;
+	}
 	
 	/* Ok, lets try some ioctls */
 
@@ -341,7 +361,7 @@ int Create(struct supertype *st, char *mddev, int mdfd,
 		array.md_minor = minor(stb.st_rdev);
 	array.not_persistent = 0;
 	/*** FIX: Need to do something about RAID-6 here ***/
-	if ( ( (level == 5) &&
+	if ( ( (level == 4 || level == 5) &&
 	       (insert_point < raiddisks || first_missing < raiddisks) )
 	     ||
 	     ( level == 6 && missing_disks == 2)
@@ -357,7 +377,7 @@ int Create(struct supertype *st, char *mddev, int mdfd,
 		 * which is array.size * raid_disks / ncopies;
 		 * .. but convert to sectors.
 		 */
-		int ncopies = (layout>>8) * (layout & 255);
+		int ncopies = ((layout>>8) & 255) * (layout & 255);
 		bitmapsize = (unsigned long long)size * raiddisks / ncopies * 2;
 /*		printf("bms=%llu as=%d rd=%d nc=%d\n", bitmapsize, size, raiddisks, ncopies);*/
 	} else
@@ -391,20 +411,33 @@ int Create(struct supertype *st, char *mddev, int mdfd,
 
 	if (name == NULL || *name == 0) {
 		/* base name on mddev */
+		/*  /dev/md0 -> 0
+		 *  /dev/md_d0 -> d0
+		 *  /dev/md/1 -> 1
+		 *  /dev/md/d1 -> d1
+		 *  /dev/md/home -> home
+		 *  /dev/mdhome -> home
+		 */
 		name = strrchr(mddev, '/');
 		if (name) {
 			name++;
-			if (strncmp(name, "md", 2)==0 &&
-			    strlen(name) > 2 &&
+			if (strncmp(name, "md_d", 4)==0 &&
+			    strlen(name) > 4 &&
+			    isdigit(name[4]) &&
 			    (name-mddev) == 5 /* /dev/ */)
+				name += 3;
+			else if (strncmp(name, "md", 2)==0 &&
+				 strlen(name) > 2 &&
+				 isdigit(name[2]) &&
+				 (name-mddev) == 5 /* /dev/ */)
 				name += 2;
 		}
 	}
-	if (!st->ss->init_super(st, &super, &array, size, name, homehost))
+	if (!st->ss->init_super(st, &super, &array, size, name, homehost, uuid))
 		return 1;
 
 	if (bitmap_file && vers < 9003) {
-		major = BITMAP_MAJOR_HOSTENDIAN;
+		major_num = BITMAP_MAJOR_HOSTENDIAN;
 #ifdef __BIG_ENDIAN
 		fprintf(stderr, Name ": Warning - bitmaps created on this kernel are not portable\n"
 			"  between different architectured.  Consider upgrading the Linux kernel.\n");
@@ -416,8 +449,9 @@ int Create(struct supertype *st, char *mddev, int mdfd,
 			fprintf(stderr, Name ": internal bitmaps not supported by this kernel.\n");
 			return 1;
 		}
-		if (!st->ss->add_internal_bitmap(st, super, bitmap_chunk, delay, write_behind,
-						 bitmapsize, 1, major)) {
+		if (!st->ss->add_internal_bitmap(st, super, &bitmap_chunk,
+						 delay, write_behind,
+						 bitmapsize, 1, major_num)) {
 			fprintf(stderr, Name ": Given bitmap chunk size not supported.\n");
 			return 1;
 		}
@@ -447,7 +481,7 @@ int Create(struct supertype *st, char *mddev, int mdfd,
 		if (CreateBitmap(bitmap_file, force, (char*)uuid, bitmap_chunk,
 				 delay, write_behind,
 				 bitmapsize,
-				 major)) {
+				 major_num)) {
 			return 1;
 		}
 		bitmap_fd = open(bitmap_file, O_RDWR);
@@ -502,6 +536,7 @@ int Create(struct supertype *st, char *mddev, int mdfd,
 				fstat(fd, &stb);
 				disk.major = major(stb.st_rdev);
 				disk.minor = minor(stb.st_rdev);
+				remove_partitions(fd);
 				close(fd);
 			}
 			switch(pass){
