@@ -121,10 +121,12 @@ int Grow_Add_device(char *devname, int fd, char *newdev)
 	info.disk.minor = minor(stb.st_rdev);
 	info.disk.raid_disk = d;
 	info.disk.state = (1 << MD_DISK_SYNC) | (1 << MD_DISK_ACTIVE);
-	st->ss->update_super(&info, super, "grow", newdev, 0, 0, NULL);
+	st->ss->update_super(&info, super, "linear-grow-new", newdev,
+			     0, 0, NULL);
 
 	if (st->ss->store_super(st, nfd, super)) {
-		fprintf(stderr, Name ": Cannot store new superblock on %s\n", newdev);
+		fprintf(stderr, Name ": Cannot store new superblock on %s\n",
+			newdev);
 		close(nfd);
 		return 1;
 	}
@@ -174,12 +176,9 @@ int Grow_Add_device(char *devname, int fd, char *newdev)
 		info.array.nr_disks = nd+1;
 		info.array.active_disks = nd+1;
 		info.array.working_disks = nd+1;
-		info.disk.number = nd;
-		info.disk.major = major(stb.st_rdev);
-		info.disk.minor = minor(stb.st_rdev);
-		info.disk.raid_disk = nd;
-		info.disk.state = (1 << MD_DISK_SYNC) | (1 << MD_DISK_ACTIVE);
-		st->ss->update_super(&info, super, "grow", dv, 0, 0, NULL);
+
+		st->ss->update_super(&info, super, "linear-grow-update", dv,
+				     0, 0, NULL);
 		
 		if (st->ss->store_super(st, fd2, super)) {
 			fprintf(stderr, Name ": Cannot store new superblock on %s\n", dv);
@@ -255,6 +254,11 @@ int Grow_addbitmap(char *devname, int fd, char *file, int chunk, int delay, int
 			devname);
 		return 1;
 	}
+	if (array.level <= 0) {
+		fprintf(stderr, Name ": Bitmaps not meaningful with level %s\n",
+			map_num(pers, array.level)?:"of this array");
+		return 1;
+	}
 	bitmapsize = array.size;
 	bitmapsize <<= 1;
 	if (get_dev_size(fd, NULL, &array_size) &&
@@ -863,8 +867,9 @@ int Grow_restart(struct supertype *st, struct mdinfo *info, int *fdlist, int cnt
 {
 	int i, j;
 	int old_disks;
-	int err = 0;
 	unsigned long long *offsets;
+	unsigned long long  nstripe, ostripe, last_block;
+	int ndata, odata;
 
 	if (info->delta_disks < 0)
 		return 1; /* cannot handle a shrink */
@@ -956,7 +961,7 @@ int Grow_restart(struct supertype *st, struct mdinfo *info, int *fdlist, int cnt
 				    fd, __le64_to_cpu(bsb.devstart)*512,
 				    0, __le64_to_cpu(bsb.length)*512)) {
 			/* didn't succeed, so giveup */
-			return -1;
+			return 1;
 		}
 
 		/* Ok, so the data is restored. Let's update those superblocks. */
@@ -975,5 +980,24 @@ int Grow_restart(struct supertype *st, struct mdinfo *info, int *fdlist, int cnt
 		/* And we are done! */
 		return 0;
 	}
-	return err;
+	/* Didn't find any backup data, try to see if any
+	 * was needed.
+	 */
+	nstripe = ostripe = 0;
+	odata = info->array.raid_disks - info->delta_disks - 1;
+	if (info->array.level == 6) odata--; /* number of data disks */
+	ndata = info->array.raid_disks - 1;
+	if (info->new_level == 6) ndata--;
+	last_block = 0;
+	while (nstripe >= ostripe) {
+		nstripe += info->new_chunk / 512;
+		last_block = nstripe * ndata;
+		ostripe = last_block / odata / (info->array.chunk_size/512) *
+			(info->array.chunk_size/512);
+	}
+
+	if (info->reshape_progress >= last_block)
+		return 0;
+	/* needed to recover critical section! */
+	return 1;
 }
